##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################


import threading
from functools import wraps
from types import FunctionType

def synchronized(method):
    @wraps(method)
    def f(*args, **kwargs):
        self = args[0]
        self.mutex.acquire();
        # print(method.__name__, 'acquired')
        try:
            return method(*args, **kwargs)
        finally:
            self.mutex.release();
            # print(method.__name__, 'released')
    return f

def dont_synchronize(f):
    #f.synchronize = False
    setattr(f, 'synchronize', False)
    return f

# check if an object should be decorated
def do_synchronize(attr, value):
    # result = ('__' not in attr and
    result = (isinstance(value, FunctionType) and
              getattr(value, 'synchronize', True))
    return result

class SynchronizedType(type):
    def __new__(cls, name, bases, namespace, **kwds):
        #not_synchronize=namespace.get('__not_synchronized__', []) 
        #synchronize=namespace.get('__synchronized__', []) 
        namespace['mutex'] = threading.RLock()
        for (method, val) in namespace.items():
            #synchronize_name=not synchronize or method in synchronize
            #not_synchronize_name=not_synchronize and method in not_synchronize
            if callable(val) and method != '__init__' and do_synchronize(method, val): #and synchronize_name and not  not_synchronize_name:
                #print("synchronizing %s" % method)
                namespace[method] = synchronized(val)
            else:
                #print("skip synchronizing %s" % method)
                pass
        return type.__new__(cls, name, bases, namespace, **kwds)   

class SynchronizeAll(metaclass=SynchronizedType):
    pass

# You can create your own self.mutex, or inherit
# from this class:
class Synchronization:
    def __init__(self):
        self.mutex = threading.RLock()       



'''
Created on Jun 29, 2016

@author: arnon
'''
import pprint
import json

class Node(object):
    def __init__(self, name, value=None):
        self._name=name
        self._value=value
        
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, n):
        self._name = n
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, v):
        self._value = v
    
    def __eq__(self, other):
        return self.name == other.name
    
    def __repr__(self):
        return repr(self.name) + ' : ' + repr(self.value)
    
    def __str__(self):
        return repr(self)
    
class Edge(object):
    def __init__(self, node1, node2):
        self.source=node1
        self.target=node2
        
    def from_node(self):
        return self.source
    
    def to_node(self):
        return self.target  
        
class DiGraph(object):
    def __init__(self, nodes=None):
        self.nodes= nodes if nodes else dict()
        #self.dfs_weights=dict()
        
    Node=Node
              
    def add_node(self, node):
        found=self.nodes.get(node.name)
        if not found:
            result=dict([('node', node),
                         ('nexts', list()),
                         ('prevs', list()),])
            self.nodes[node.name]=result
        return found is None          
        
    def add_edge(self, from_node, *to_nodes):
        fname=from_node.name
        fnode=self.nodes.get(fname)
        if not fnode: 
            raise Exception("EdgeError: From-Node {} not in tree".format(from_node.name))
            #fnode=self.Node(from_node)
            #self.add_node(from_node)
        
        nexts=fnode['nexts']
        for node in to_nodes:
            tname=node.name
            tnode=self.nodes.get(node.name)
            if not tnode: 
                raise Exception("EdgeError: To-Node {} not in tree".format(node.name))
            
            if tname not in nexts:
                nexts.append(tname)
            prevs=tnode['prevs']
            if fname not in prevs:
                prevs.append(fname)
            #print('Added Edge:', from_node, node)
    
    def del_edge(self, from_node, *to_nodes):
        fname=from_node.name
        fnode=self.nodes.get(fname)
        nexts=fnode['nexts']
        for node in to_nodes:
            tname=node.name
            try: nexts.remove(tname)
            except: pass
            
            tnode=self.nodes.get(tname)
            prevs=tnode['prevs']
            try: prevs.remove(fname)
            except: pass
    
    def del_node(self, node):
        dname=node.name
        dnode=self.nodes.get(dname)
        nexts=dnode['nexts']
        for tname in nexts:
            tnode=self.nodes.get(tname)
            prevs=tnode['prevs']
            try: prevs.remove(dname)
            except: pass
            
        prevs=dnode['prevs']
        for fname in prevs:
            fnode=self.nodes.get(fname)
            nexts=fnode['nexts']
            try: nexts.remove(dname)
            except: pass
        del self.nodes[dname]
    
    def _dfs(self, node_name, weights):
        props=self.nodes[node_name]
        try:
            for name in props['nexts']:
                if name not in weights.keys():
                    self._dfs(name, weights)
            weight=1+max([weights[name] for name in props['nexts']])
        except:
            weight=1
        weights[node_name]=weight
        
    def __len__(self):
        return len(self.nodes)
        
        
    def dfs(self, reverse=False):
        ''' Node iterator sorted depth first '''
        weights=dict()
        for node_name, props in self.nodes.items():
            if node_name not in weights.keys():
                self._dfs(node_name, weights)
        for name in sorted(weights.keys(), key=lambda x: weights[x], reverse=not reverse):
            yield self.nodes[name]['node']
     
    def __str__(self):
        return pprint.pformat(self.nodes) 

    def to_file(self, fp):
        json.dump(self.nodes, fp)
    
    @classmethod
    def from_file(cls, fp):
        try:
            nodes=json.load(fp)
        except:
            return None
        else:
            return cls(nodes)
 
        
if __name__ == '__main__':
    g=DiGraph()
   
    a=g.Node(('a', 1), 35)
    b=g.Node(('b', 2), 46)
    c=g.Node(('c', 3), 78)
    d=g.Node(('d', 4), 89)
    
    g.add_node(a)  
    g.add_node(b)  
    g.add_node(c)  
    g.add_node(d) 
       
    g.add_edge(a, b)
    g.add_edge(b, c)
    g.add_edge(c, d)
   
    print(g)
    
    g.del_node(d) 

    print(g)
    
    for node in g.dfs(reverse=True):
        print(node.name, node.value)
        


'''
Created on Oct 11, 2017

@author: arnon
'''

class LogSim(object):
    def __init__(self, file=None):
        self.file = file
        if file:
            self.file = open(file, 'w')
    
    def debug(self, msg):
        self._print('debug', msg)
    
    def info(self, msg):
        self._print('info', msg)
    
    def warning(self, msg):
        self._print('warning', msg)
    
    def error(self, msg):
        self._print('error', msg)
    
    def critical(self, msg):
        self._print('critical', msg)
        
    def _print(self, key, msg):
        print("{}: {}".format(key, msg), file=self.file)
        
    def close(self):
        if self.file:
            self.file.close()
            
    def __del__(self):
        self.close()




# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################


from threading import Lock


class SingletonType(type):
    '''
        Thread friendly Singleton construct
    '''
    __locker=Lock() 
    __instance={}
    
    def __call__( self, name='', *args, **kwargs):   
        SingletonType.__locker.acquire()
        try:
            instance=self.__instance[name]
        except KeyError:
            instance=super(SingletonType, self).__call__(*args, **kwargs)
            self.__instance[name]=instance
        SingletonType.__locker.release()
        return instance
    
        
class Singleton(metaclass=SingletonType):
    pass

class NamedSingletonType(type):
    '''
        Thread friendly Singleton construct
    '''
    __locker=Lock() 
    __instance={}
      
    def __call__( self, name='', *args, **kwargs):   
        NamedSingletonType.__locker.acquire()
        try:
            instance=self.__instance[name]
        except KeyError:
            instance=super(NamedSingletonType, self).__call__(*args, name=name, **kwargs)
            self.__instance[name]=instance
        NamedSingletonType.__locker.release()
        return instance
    
        
class NamedSingleton(metaclass=NamedSingletonType):
    pass

if __name__ == '__main__':

    class SingTest(Singleton):
        __data=None
        
        def __init__(self, mydata=None, *args, **kwargs):
            print('init', mydata, args, kwargs)
            self.__data = mydata
        
        def load(self, mydata=None):
            if not self.__data:
                self.__data = mydata
            return self
                
        def get(self):
            return self.__data
        
    
    s1=SingTest('S1', 55, a=23).load(1)
    print('s1', s1.get())
    
    s2=SingTest('S2')
    print('s2', s2.get())
    s2.load(2)
    print('s2', s2.get())
    s2.load(3)
    print('s2', s2.get())
    
    print('s1', s1.get())
 
    class NamedSingTest(NamedSingleton):
        __data=None
        
        def __init__(self, mydata=None, name='', *args, **kwargs):
            print('init', mydata, name, args, kwargs)
            self.name=name
            self.__data = mydata
        
        def load(self, mydata=None):
            if not self.__data:
                self.__data = mydata
            return self
                
        def get(self):
            return self.__data
        
    
    s1=NamedSingTest('S1', 55, a=23).load(1)
    print('s1', 'named', s1.name, s1.get())
    
    s2=NamedSingTest('S2')
    print('s2', 'named', s2.name, s2.get())
    s2.load(2)
    print('s2', 'named', s2.name, s2.get())
    s2.load(3)
    print('s2', 'named', s2.name, s2.get())
    
    print('s1', 'named', s1.name, s1.get())

    class Sequence(NamedSingleton):
        step_id=0
        
        def __init__(self, name=''):
            self.name=name

        def __call__(self,):
            step_id=self.step_id
            self.step_id += 1
            return step_id  

    A=Sequence('A')
    print(A.name, A())
    print(A.name, A())
    B=Sequence('B')
    print(B.name, B()) 
           
    


'''
Created on Dec 8, 2016

@author: arnon
'''
import datetime
from types import FunctionType
import inspect
from functools import wraps

# check if an object should be decorated
def do_decorate(attr, value):
    # result = ('__' not in attr and
    result = (isinstance(value, FunctionType) and
              getattr(value, 'decorate', True))
    return result

# decorate all instance methods (unless excluded) with the same decorator
def decorated_class(decorator):
    class DecorateAll(type):
        def __new__(cls, name, bases, dct):
            for attr, value in dct.items():
                if do_decorate(attr, value):
                    decorated=decorator(name, value)
                    dct[attr] = decorated
                    #print('decorated', attr, decorated)
            return super(DecorateAll, cls).__new__(cls, name, bases, dct)
    return DecorateAll

# decorator to exclude methods
def dont_decorate(f):
    f.decorate = False
    return f

def traced_method(print_func=None, print_args=False, print_result=False):
    ''' provides decorator for method or function to log entry and exit
    
    Args:
        print_func: method to use to print information. defaults to print_func
        print_args: if set, would add representation of arguments
        print_result: if set, would add representation of returned value
        
    Returns:
        decorator
    '''
    def print_method_name(name, f=None):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        caller="%s.%s(%s)" % (filename, function_name, line_number)
        if f is None :
            text_id='%s' % (name.__name__)
            method=name
            is_method=True
        else:
            text_id='%s.%s' % (name, f.__name__, )
            method=f
            is_method=False
            
        def wrapper(*args, **kwargs):
            
            start=datetime.datetime.now()
            func_args=func_result=''
            if print_args:
                show_args=args if not is_method else args[1:]
                func_args="[ args: %s ][ kwargs: %s ]" % (show_args, kwargs)
            if print_func:
                #print_func('[ %s ][ %s ][ entering]%s[ %s ]' % (str(start), text_id, func_args,caller))
                print_func('[ %s ][ entering]%s[ %s ]' % (str(text_id), str(func_args), str(caller)))
            result=method(*args, **kwargs)
            #print_func('Result: %s' % (repr(result),))
            #print(str(result))
            finish=datetime.datetime.now()
            if print_result:
                func_result="[ result: %s ]" % repr(result)
            if print_func:
                #print_func('Result: %s' % (repr(result),))
                #print_func('[ %s ][ %s ][ exiting ] [ time span: %s][ %s ]' % (str(finish), text_id, str(finish-start), caller))
                print_func('[ %s ][ exiting ] [ time span: %s]%s[ %s ]' % (str(text_id), str(finish-start), str(func_result), str(caller)))
            return result
        return wrapper
    return print_method_name

class TracedMethod(object):
    def __init__(self, print_func=print, print_args=False):
        self.print_func=print_func
        self.print_args=print_args
        
    def __call__(self, name, f=None):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        caller="%s.%s(%s)" % (filename, function_name, line_number)
        if f==None :
            text_id='%s' % (name.__name__)
            method=name
        else:
            text_id='%s.%s' % (name, f.__name__, )
            method=f

        @wraps(method)
        def wrapper(*args, **kwargs):
            
            start=datetime.datetime.now()
            func_args=''
            if self.print_args:
                func_args="[ args: %s ][ kwargs: %s ]" % (args, kwargs)
            if self.print_func:
                #print_func('[ %s ][ %s ][ entering]%s[ %s ]' % (str(start), text_id, func_args,caller))
                self.print_func('[ %s ][ entering]%s[ %s ]' % (text_id, func_args,caller))
            result=method(*args, **kwargs)
            #print_func('Result: %s' % (repr(result),))
            #print(str(result))
            finish=datetime.datetime.now()
            if self.print_func:
                #print_func('Result: %s' % (repr(result),))
                #print_func('[ %s ][ %s ][ exiting ] [ time span: %s][ %s ]' % (str(finish), text_id, str(finish-start), caller))
                self.print_func('[ %s ][ exiting ] [ time span: %s][ %s ]' % (text_id, str(finish-start), caller))
            return result
        return wrapper

class LogCaller(object):
    def __init__(self, msg='', logger=None):
        self.msg=msg if not msg else "%s;" % msg
        self.show=logger.debug if logger is not None else print
        
    def __call__(self, name, func=None):
        (frame, filename, line_number,
         function_name, lines, index) = inspect.getouterframes(inspect.currentframe())[1]
        caller="%s.%s(%s)" % (filename, function_name, line_number)
        
        self.show("%s %s" % (self.msg, caller)) 
        #print("Showing....", self.msg, caller)
        if func==None:
            method=name
        else:
            method=func
        
        @wraps(method)
        def wrapper(*args, **kwargs):
            result=method(*args, **kwargs)
            return result
        return wrapper


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

from logging import INFO, Handler
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
import os

class TimedSizedRotatingHandler(TimedRotatingFileHandler, RotatingFileHandler):
    
    def __init__(self, filename, file_mode='a', maxBytes=0, backupCount=0, encoding='ascii', delay=False, when='h', interval=1, utc=False, atTime=None, *args, **kwargs):
        """ 
        Combines RotatingFileHandler TimedRotatingFileHandler)  
        """
        
        if file_mode.startswith('w'):
            try: os.remove(filename)
            except: pass
        self.filename=filename
        RotatingFileHandler.__init__(self, filename=filename, mode=file_mode, maxBytes=maxBytes, backupCount=backupCount, encoding=encoding, delay=delay)
        TimedRotatingFileHandler.__init__(self, filename=filename, when=when, interval=interval, backupCount=backupCount, encoding=encoding, delay=delay, utc=utc, atTime=atTime)
        
    def shouldRollover(self, record):
        """
        Check the need to rotate.     
        """
        timed_rollover = TimedRotatingFileHandler.shouldRollover(self, record) 
        sized_rollover = RotatingFileHandler.shouldRollover(self, record)
        
        return timed_rollover or sized_rollover

    def doRollover(self):
        """
        It is enough to use timed base rollover.
        """
        super(TimedRotatingFileHandler, self).doRollover()

    def getFilesToDelete(self):
        """
        It is enough to use timed base rollover.
        """
        return super(TimedRotatingFileHandler, self).getFilesToDelete()        



def get_file_handler(logdir='', name=None, formatter=None, file_prefix=None, file_suffix=None, *args, **kwargs):
    '''
    
    Args:
        kwargs:
            file_mode='a', 
            maxBytes=0, 
            backupCount=0, 
            encoding='ascii', 
            delay=False, 
            when='h', 
            interval=1, 
            utc=False, 
            atTime=None
    '''
    result=list()
    if logdir is None: logdir=''
    
    if logdir:
        if not os.path.isdir(logdir):
            try:
                os.makedirs(logdir, mode=0o744, exist_ok=True)
            except:
                raise

    #key_s=''
    if file_suffix: name = "%s.%s" %(name, file_suffix)
    if name: name = "%s.log" % name
    else: name = 'logger.log'
    if file_prefix: name = "%s.%s" %(file_prefix, name)
    
    #print('get_file_handlers: process_key:', process_key)
    #traceback.print_stack()
    filename = os.path.join(logdir, name)
    #del kwargs['delay']
    handler = TimedSizedRotatingHandler(filename=filename, **kwargs)
    #handler.setLevel(logging_level)
    #formatter = formatter if formatter is not None else LevelBasedFormatter(level_formats=level_formats, datefmt=datefmt)
    #print('get_file_handler formatter:', formatter)
    if formatter is None: raise
    handler.setFormatter(formatter)
    result.append(handler)
    # create error file handler and set level to error
    
    return result

class HierarchicalTimedSizedRotatingHandler(Handler):
    def __init__(self, key='name', separator='.', consolidate='', *args, **kwargs):
        ''' Maintains TimedSizedRotatingHandler handlers according to hierarchy.
        
        LogRecord key will be used to store and fetch handlers.
        If there is separator, value of LogRecord key will be split from the right generating multiple values.
        E.g., A.B.C value with be translated to hierarchical values of A, A.B, and A.B.C.
        Note, it is important that LogRecord key values will not include separator characters.
        
        Args:
            key: name of LogRecord attributes to use to associate handlers.
            separator: is used to split key to hierarchy. If None, key value will not be broken to parts.
            consolidate: name to which records will be consolidated into.
            
            kwargs: 
                logdir='', 
                name=None, 
                formatter=None, 
                file_prefix=None, 
                file_suffix=None
                
                filename, 
                file_mode='a', 
                maxBytes=0, 
                backupCount=0, 
                encoding='ascii', 
                delay=False, 
                when='h', 
                interval=1, 
                utc=False, 
                atTime=None
        '''
        super(HierarchicalTimedSizedRotatingHandler, self).__init__() 

        self.separator = separator
        self.key = key
        self.consolidate = consolidate
        self.handler_args = args
        self.handler_kwargs = kwargs
        self.__handlers = dict()
        
    def handle(self, record):
        """ Override handle a record.  Loops through handlers according to key hierarchy offering them the record
        to handle.

        Args:
            record: The record to handle.
        """
        
        # Find handlers that match process keys
        
        #attributes = record.__dict__
        #record_name = attributes.get('name', None)
        record_name = getattr(record, 'name')
        #for process_key in self.process_key:
        #record_key = attributes.get(self.key, None)
        try: record_key = getattr(record, self.key)
        except KeyError: record_key = ''
        #print('record_key[process_key]: %s[%s]' %(record_key, self.key))
        #print('record_key[processName]: %s' %(repr(record.__dict__)))
        
        keys = [record_key]
        if self.consolidate: keys.append(self.consolidate)
        if self.separator is not None:
            keys = []
            left_key = record_key
            while left_key:
                keys.append(left_key)
                left_key = left_key.rpartition(self.separator)[0]
        
        #print('consolidate keys %s: %s' % (record_key, keys))
        handlers = list()
        for record_key in keys:
            #if record_key: 
            #process_handlers = self.handlers[process_key]
            key_handlers = self.__handlers.get(record_key,)
            # avoid getting dedicated handler when in consolidated mode and record with 
            # name equal to the global.
            need_handler = key_handlers is None or record_key != self.name 
            if need_handler:
                key_handlers = get_file_handler(*self.handler_args, name=record_key, **self.handler_kwargs)
                self.__handlers[record_key] = key_handlers
                
            handlers.extend(key_handlers)
                
        #if len(self.global_handlers) > 0:
        #    handlers.extend(self.global_handlers)
            
        #if len(self.console_handlers) > 0:
        #    handlers.extend(self.console_handlers)
        
        record = self.prepare(record)
        
        for handler in list(set(handlers)):
            if record.levelno >= handler.level: # This check is not in the parent class
                handler.handle(record)
                
    def prepare(self, record):
        return record

    def addConsoleHandler(self, handler):
        self.console_handlers.append(handler)
        
    def addHandler(self, handler):
        """
        Add the specified handler to this logger.
        
        handler is expected to have process_key attribute.
        process_key attribute is expected to be a list of records attribute names that handler would bind to.
        if handler does not have process_key attribute or it is empty, handler will be associated with 
        """
        key_bind=False
        if hasattr(handler, 'process_key'):
            handler_key=handler.process_key
            for key in list(set(self.process_key) & set(handler_key)):
                exist_handler=self.key_handlers.get(key, list())
                self.key_handlers[key]=exist_handler
                exist_handler.append(handler)
                key_bind=True
        #if not key_bind:
        #    self.global_handlers.append(handler)

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this logger.
        """
        if hdlr in self.handlers:
            hdlr.close()
            self.handlers.remove(hdlr)
        
        
        
        
        


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
from .synchronize import SynchronizeAll


class Observer(object):
    def update(self, observable, arg):
        '''Called when the observed object is
        modified. You call an Observable object's
        notifyObservers method to notify all the
        object's observers of the change.'''
        pass

class Observable(SynchronizeAll):
    __synchronized__=["addObserver", "deleteObserver", "deleteObservers", 
                     "setChanged", "clearChanged", "hasChanged", "countObservers"]
    def __init__(self):
        self.obs = []
        self.changed = 0
        super(Observable, self).__init__()

    def addObserver(self, observer):
        if observer not in self.obs:
            self.obs.append(observer)

    def deleteObserver(self, observer):
        self.obs.remove(observer)

    def notifyObservers(self, arg = None):
        '''If 'changed' indicates that this object
        has changed, notify all its observers, then
        call clearChanged(). Each observer has its
        update() called with two arguments: this
        observable object and the generic 'arg'.'''

        self.mutex.acquire()
        try:
            if not self.changed: return
            # Make a local copy in case of synchronous
            # additions of observers:
            localArray = self.obs[:]
            self.clearChanged()
        finally:
            self.mutex.release()
        # Updating is not required to be synchronized:
        for observer in localArray:
            observer.update(self, arg)

    def deleteObservers(self): self.obs = []
    def setChanged(self): self.changed = 1
    def clearChanged(self): self.changed = 0
    def hasChanged(self): return self.changed
    def countObservers(self): return len(self.obs)

if __name__ == '__main__':
    # Observer/ObservedFlower.py
    # Demonstration of "observer" pattern.
        
    class Flower:
        def __init__(self):
            self.isOpen = 0
            self.openNotifier = Flower.OpenNotifier(self)
            self.closeNotifier= Flower.CloseNotifier(self)
        def open(self): # Opens its petals
            self.isOpen = 1
            self.openNotifier.notifyObservers()
            self.closeNotifier.open()
        def close(self): # Closes its petals
            self.isOpen = 0
            self.closeNotifier.notifyObservers()
            self.openNotifier.close()
        def closing(self): return self.closeNotifier
    
        class OpenNotifier(Observable):
            def __init__(self, outer):
                Observable.__init__(self)
                self.outer = outer
                self.alreadyOpen = 0
                
            def notifyObservers(self):
                if self.outer.isOpen and not self.alreadyOpen:
                    self.setChanged()
                    Observable.notifyObservers(self)
                    self.alreadyOpen = 1
                    
            def close(self):
                self.alreadyOpen = 0
    
        class CloseNotifier(Observable):
            def __init__(self, outer):
                Observable.__init__(self)
                self.outer = outer
                self.alreadyClosed = 0
            def notifyObservers(self):
                if not self.outer.isOpen and not self.alreadyClosed:
                    self.setChanged()
                    Observable.notifyObservers(self)
                    self.alreadyClosed = 1
            def open(self):
                self.alreadyClosed = 0
    
    class Bee:
        def __init__(self, name):
            self.name = name
            self.openObserver = Bee.OpenObserver(self)
            self.closeObserver = Bee.CloseObserver(self)
        # An inner class for observing openings:
        class OpenObserver(Observer):
            def __init__(self, outer):
                self.outer = outer
            def update(self, observable, arg):
                print("Bee " + self.outer.name + "'s breakfast time!")
        # Another inner class for closings:
        class CloseObserver(Observer):
            def __init__(self, outer):
                self.outer = outer
            def update(self, observable, arg):
                print("Bee " + self.outer.name + "'s bed time!")
    
    class Hummingbird:
        def __init__(self, name):
            self.name = name
            self.openObserver = Hummingbird.OpenObserver(self)
            self.closeObserver = Hummingbird.CloseObserver(self)
        class OpenObserver(Observer):
            def __init__(self, outer):
                self.outer = outer
            def update(self, observable, arg):
                print("Hummingbird " + self.outer.name + "'s breakfast time!")
        class CloseObserver(Observer):
            def __init__(self, outer):
                self.outer = outer
            def update(self, observable, arg):
                print("Hummingbird " + self.outer.name + "'s bed time!")
    
    f = Flower()
    ba = Bee("Eric")
    bb = Bee("Eric 0.5")
    ha = Hummingbird("A")
    hb = Hummingbird("B")
    
    f.openNotifier.addObserver(ha.openObserver)
    f.closeNotifier.addObserver(ha.closeObserver)
    
    f.openNotifier.addObserver(hb.openObserver)
    f.closeNotifier.addObserver(hb.closeObserver)
    
    f.openNotifier.addObserver(ba.openObserver)
    f.closeNotifier.addObserver(ba.closeObserver)
    
    f.openNotifier.addObserver(bb.openObserver)
    f.closeNotifier.addObserver(bb.closeObserver)
    
    # Hummingbird 2 decides to sleep in:
    f.openNotifier.deleteObserver(hb.openObserver)
    # A change that interests observers:
    f.open()
    f.open() # It's already open, no change.
    # Bee 1 doesn't want to go to bed:
    f.closeNotifier.deleteObserver(ba.closeObserver)
    f.close()
    f.close() # It's already closed; no change
    f.openNotifier.deleteObservers()
    f.open()
    f.close()


'''
Created on Sep 19, 2017

@author: arnon
'''
import socket
from copy import deepcopy

LOCAL_HOST = '127.0.0.1'
def get_free_port():
    
    s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind((LOCAL_HOST,0))
    host,port=s.getsockname()
    s.close()
    return port


def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("8.8.8.8", 80))
    except OSError:
        address = 'localhost'
    else:
        address = s.getsockname()[0]
    s.close()
    return address


def get_hostname(full=False):
    ip = get_ip_address()
    if full == False:
        try:
            name = socket.gethostbyaddr(ip)[0]
        except:
            name = ip
    else:
        name = socket.getfqdn(ip)
    return name


def hostname_resolves(hostname):
    try:
        socket.gethostbyname(hostname)
        return 1
    except socket.error:
        return 0


if __name__ == '__main__':
    print(get_free_port())
    print(get_ip_address())
    print(get_hostname())
    
    


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

"""
About
=========
:synopsis:     threaded provides decorator for class methods to run via a thread
:moduleauthor: Arnon Sela
:date:         2016/10/28
:description:  a method decorated with @threaded would run in a separated thread. this work 
   
:comment: based on http://code.activestate.com/recipes/576684-simple-threading-decorator/



    
"""


from functools import wraps
from threading import Thread, RLock, Event #, current_thread
import logging

logger=logging.getLogger(__name__)

__all__ = ['threaded']

from .decorate import traced_method
traced=traced_method(print, print_args=False, print_result=False)

class ThreadedResult(object): 
    """Represents an asynchronous operation that may not have completed yet.""" 
    def __init__(self): 
        self.completed = False 
        self.failed = False 
        self.__wait = Event() 
        self.__callbacks = [] 
        self.__errbacks = [] 
        self.__retval = None 
        self.__error = None 
        self.__lock = RLock()
        self.name=None
        self.running=False
        self.thread=None
        
    def __release(self):
        self.__lock.release()
    
    def __acquire(self):
        self.__lock.acquire()
    
    def start(self):
        self.running=True
        
    def started(self):
        return self.running
    
    def set_thread(self, t):
        self.thread=t
        
    def join(self, timeout=None):
        if self.thread:
            self.thread.join(timeout)

    def completed(self):
        return self.completed

    def complete(self):
        self.__acquire()
        self.completed = True
        self.running = False
        self.__wait.set()
        self.__release()
    
    def succeed(self, retval):
        self.__retval = retval
        self.complete()
        for callback in self.__callbacks:
            callback(retval)
        self.clearCallbacks()
    
    def fail(self, error):
        self.__error = error
        self.failed = True
        self.complete()
        for errback in self.__errbacks:
            errback(error)
        self.clearCallbacks()
    
    def clearCallbacks(self):
        self.__callbacks = []
        self.__errbacks = []
    
    def addCallback(self, callback,):
        self.__acquire()
        #self.name=callback.name
        try:
            if self.completed:
                if not self.failed:
                    callback(self.__retval)
            else:
                self.__callbacks.append(callback)
        finally:
            self.__release()
    
    def addErrback(self, errback):
        self.__acquire()
        self.name=errback.name
        try:
            if self.completed:
                if self.failed:
                    errback(self.__error)
            else:
                self.__errbacks.append(errback)
        finally:
            self.__release()
    
    def __getResult(self):
        self.__wait.wait()
        if not self.failed:
            return self.__retval
        else:
            raise self.__error
    result=property(__getResult)
    
    def syncResult(self):
        return self.__getResult()


class RetriveAsycValue(object):
    def __init__(self, name):
        self.name=name
        self.result=None
        
    def __call__(self, retval):
        self.result=retval 
    
import inspect

def traces(trace, start=0, end=None):
    '''
     File "/private/var/acrisel/sand/gradior/gradior/gradior/gradior/loop_task.py", line 41, in task_wrapper
    '''
    result=[ "File \"%s\", line %s, in %s\n    %s" % (frame.filename, frame.lineno, frame.function, frame.code_context[0].rstrip()) for frame in trace]
    return result
   
def threaded(method): 
    @wraps(method)
    def wrapper(*args, **kwargs): 
        treated_result = ThreadedResult() 
        def _method(): 
            treated_result.start()
            try: 
                result=method(*args, **kwargs) 
            except Exception as e: 
                trace=inspect.trace()
                trace=traces(trace, 2)
                treated_result.fail(trace)
            else:
                treated_result.succeed(result)
        t=Thread(target = _method)
        t.start() 
        treated_result.set_thread(t)
        return treated_result 
    return wrapper


class Threaded(object):
    def __init__(self, log=True):
        self.log=log
        
    def __call__(self,method):
        @wraps(method)
        def wrapper(*args, **kwargs): 
            async_result = ThreadedResult() 
            def _method(): 
                async_result.start()
                try: 
                    result=method(*args, **kwargs) 
                except Exception as e: 
                    trace=inspect.trace()
                    trace=traces(trace,start=0, end=None)
                    if self.log: logger.error("%s\n    %s" % (repr(e), '\n    '.join(trace)))
                    async_result.fail(trace)
                else:
                    async_result.succeed(result)
            t=Thread(target = _method)
            t.start() 
            async_result.set_thread(t)
            return async_result 
        return wrapper
        
    
    
    


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

class  Mediator(object):
    def __init__(self, generator):
        self.generator=generator
        self.__next_values=list()
    
    def __iter__(self):
        return self
        
    def __next__(self):
        if len(self.__next_values) >0:
            value=self.__next_values[0]
            self.__next_values=self.__next_values[1:]
            return value
        else:
            try:
                value=next(self.generator)
            except StopIteration:
                raise
            else:
                return value
        
    def has_next(self, count=1):
        if len(self.__next_values) > count:
            return True
        
        for _ in range(count-len(self.__next_values)):
            try:
                value=next(self.generator)
            except StopIteration:
                break
            else:
                self.__next_values.append(value)

        result=len(self.__next_values) - count >= 0
        return result
    



'''
Created on Sep 21, 2017

@author: arnon
'''
import logging
from copy import copy
from datetime import datetime
from acrilib import get_hostname, get_ip_address

class LoggerAddHostFilter(logging.Filter):
    """
    This is filter adds host information to LogRecord.
    """

    def filter(self, record):

        if not hasattr(record, 'host'):
            record.host = get_hostname()
        if not hasattr(record, 'host'):
            record.ip = get_ip_address()
        return True

class MicrosecondsDatetimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        ct = datetime.fromtimestamp(record.created)
        
        if ct is  None:
            ct=datetime.now()
            
        if datefmt is not None:
            s = ct.strftime(datefmt)
        else:
            #print('MicrosecondsDatetimeFormatter:', repr(ct),)
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s.%03d" % (t, record.msecs)
            
        return s

class LevelBasedFormatter(logging.Formatter):
    
    defaults={
        logging.DEBUG : u"%(asctime)-15s: %(process)-7s: %(levelname)-7s: %(message)s: %(module)s.%(funcName)s(%(lineno)d)",
        'default' : u"%(asctime)-15s: %(process)-7s: %(levelname)-7s: %(message)s",
        }
 
    def __init__(self, level_formats={}, datefmt=None):
        super(LevelBasedFormatter, self).__init__()
        formats = LevelBasedFormatter.defaults
        if level_formats:
            formats = copy(LevelBasedFormatter.defaults)
            formats.update(level_formats)
            
        self.datefmt = datefmt  
        self.formats = dict([(level, MicrosecondsDatetimeFormatter(fmt=fmt, datefmt=self.datefmt)) for level, fmt in formats.items()])
        self.default_format = self.formats['default']  
        #super(LevelBasedFormatter, self).__init__() # fmt=self.default_format, datefmt=self.datefmt)

    def format(self, record):
        formatter = self.formats.get(record.levelno, self.default_format,)
        #print('LevelBasedFormatter:', formatter)
        result = formatter.format(record)
        return result



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import unittest
import functools 

class MergedChainedDict(dict):
    def __init__(self, *args, submerge=False):
        ''' Merge list of dicts.  
        Merge is such that significant (left) wins over least significant (right).
        
        Args: 
            args: list of dicts where left most is most significant.
            submerge: if an entity in a dictionary is dictionary, merge-chain that entity.
        '''
        super().__init__()
        args=list(args)
        if not submerge:
            for arg in args[::-1]:
                self.update(arg)
        else:
            keys = functools.reduce(lambda x,y: x+y, [list(arg.keys()) for arg in args], list())
            for key in set(keys):
                values = [arg.get(key) for arg in args]
                values = [value for value in values if value is not None]
                if len(values) < 1: 
                    value = None
                else:
                    value0 = values[0] 
                    if not isinstance(value0, dict):
                        value = value0
                    else: # this is a dict
                        value = MergedChainedDict(*values, submerge=True)
                self[key] = value

class TestMergedChainedDict(unittest.TestCase):

    def test_uni_depth(self):
        a = {1:11, 2:22}
        b = {3:33, 4:44}
        c = {1:55, 4:66}
        d = MergedChainedDict(c, b, a)
        d_expected = {1: 55, 2: 22, 3: 33, 4: 66}
        self.assertEqual(d, d_expected)
            
    def test_two_depth_submerge(self):
        a = {1:11, 2:22, 3:{'a': 1, 'b':2, 'c':5}}
        b = {3:{'a':2, 'c': 7}, 4:44}
        c = {1:55, 4:66, 3:{'a': 3, 'b':2}}
        d = MergedChainedDict(c, b, a, submerge=True)
        d_expected = {1: 55, 2: 22, 3: {'a': 3, 'b': 2, 'c': 7}, 4: 66}
        self.assertEqual(d, d_expected)
            
if __name__ == '__main__':
    
    unittest.main()


from .VERSION import __version__

from .idioms.decorate import traced_method
from .idioms.threaded import Threaded, RetriveAsycValue, threaded
from .idioms.singleton import Singleton, NamedSingleton
from .idioms.sequence import Sequence
from .idioms.data_types import MergedChainedDict
from .idioms.synchronize import Synchronization, SynchronizeAll, dont_synchronize
from .idioms.mediator import Mediator

from .idioms.osutil import get_free_port, get_ip_address, get_hostname, hostname_resolves

from .logging.formatters import LoggerAddHostFilter, LevelBasedFormatter, MicrosecondsDatetimeFormatter
from .logging.handlers import TimedSizedRotatingHandler, HierarchicalTimedSizedRotatingHandler, get_file_handler
from .logging.logsim import LogSim